# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2026 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for email settings
"""

import colander

from wuttaweb.views import MasterView
from wuttaweb.forms.schema import EmailRecipients


class EmailSettingView(MasterView):  # pylint: disable=abstract-method
    """
    Master view for :term:`email settings <email setting>`.
    """

    model_name = "email_setting"
    model_key = "key"
    model_title = "Email Setting"
    url_prefix = "/email/settings"
    filterable = False
    sortable = True
    sort_on_backend = False
    paginated = False
    creatable = False
    deletable = False

    labels = {
        "key": "Email Key",
        "replyto": "Reply-To",
    }

    grid_columns = [
        "key",
        "subject",
        "to",
        "enabled",
    ]

    # TODO: why does this not work?
    sort_defaults = "key"

    form_fields = [
        "key",
        "fallback_key",
        "description",
        "subject",
        "sender",
        "replyto",
        "to",
        "cc",
        "bcc",
        "notes",
        "enabled",
    ]

    def __init__(self, request, context=None):
        super().__init__(request, context=context)
        self.email_handler = self.app.get_email_handler()

    def get_grid_data(self, columns=None, session=None):
        """
        This view calls
        :meth:`~wuttjamaican:wuttjamaican.email.EmailHandler.get_email_settings()`
        on the :attr:`email_handler` to obtain its grid data.
        """
        data = []
        for setting in self.email_handler.get_email_settings().values():
            data.append(self.normalize_setting(setting))
        return data

    def normalize_setting(self, setting):  # pylint: disable=empty-docstring
        """ """
        key = setting.__name__
        setting = setting(self.config)
        return {
            "key": key,
            "fallback_key": setting.fallback_key or "",
            "description": setting.get_description() or "",
            "subject": self.email_handler.get_auto_subject(
                key, rendered=False, setting=setting
            ),
            "sender": self.email_handler.get_auto_sender(key),
            "replyto": self.email_handler.get_auto_replyto(key) or colander.null,
            "to": self.email_handler.get_auto_to(key),
            "cc": self.email_handler.get_auto_cc(key),
            "bcc": self.email_handler.get_auto_bcc(key),
            "notes": self.email_handler.get_notes(key) or colander.null,
            "enabled": self.email_handler.is_enabled(key),
        }

    def configure_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_grid(g)

        # key
        g.set_searchable("key")
        g.set_link("key")

        # subject
        g.set_searchable("subject")
        g.set_link("subject")

        # to
        g.set_renderer("to", self.render_to_short)

        # enabled
        g.set_renderer("enabled", "boolean")

    def render_to_short(  # pylint: disable=empty-docstring,unused-argument
        self, setting, field, value
    ):
        """ """
        recips = value
        if not recips:
            return None

        if len(recips) < 3:
            return ", ".join(recips)

        recips = ", ".join(recips[:2])
        return f"{recips}, ..."

    def get_instance(  # pylint: disable=empty-docstring,arguments-differ,unused-argument
        self, **kwargs
    ):
        """ """
        key = self.request.matchdict["key"]
        setting = self.email_handler.get_email_setting(key, instance=False)
        if setting:
            return self.normalize_setting(setting)

        raise self.notfound()

    def get_instance_title(self, instance):  # pylint: disable=empty-docstring
        """ """
        setting = instance
        return setting["subject"]

    def configure_form(self, form):  # pylint: disable=empty-docstring
        """ """
        f = form
        super().configure_form(f)

        # fallback_key
        f.set_readonly("fallback_key")

        # description
        f.set_readonly("description")
        f.set_widget("description", "notes")

        # replyto
        f.set_required("replyto", False)

        # to
        f.set_node("to", EmailRecipients())

        # cc
        f.set_node("cc", EmailRecipients())

        # bcc
        f.set_node("bcc", EmailRecipients())

        # notes
        f.set_widget("notes", "notes")
        f.set_required("notes", False)

        # enabled
        f.set_node("enabled", colander.Boolean())

    def persist(  # pylint: disable=too-many-branches,empty-docstring,arguments-differ,unused-argument
        self, setting, **kwargs
    ):
        """ """
        session = self.Session()
        key = self.request.matchdict["key"]

        def save(name, value):
            self.app.save_setting(
                session, f"{self.config.appname}.email.{key}.{name}", value
            )

        def delete(name):
            self.app.delete_setting(
                session, f"{self.config.appname}.email.{key}.{name}"
            )

        # subject
        if setting["subject"]:
            save("subject", setting["subject"])
        else:
            delete("subject")

        # sender
        if setting["sender"]:
            save("sender", setting["sender"])
        else:
            delete("sender")

        # replyto
        if setting["replyto"]:
            save("replyto", setting["replyto"])
        else:
            delete("replyto")

        # to
        if setting["to"]:
            save("to", setting["to"])
        else:
            delete("to")

        # cc
        if setting["cc"]:
            save("cc", setting["cc"])
        else:
            delete("cc")

        # bcc
        if setting["bcc"]:
            save("bcc", setting["bcc"])
        else:
            delete("bcc")

        # notes
        if setting["notes"]:
            save("notes", setting["notes"])
        else:
            delete("notes")

        # enabled
        save("enabled", "true" if setting["enabled"] else "false")

    def render_to_response(self, template, context):  # pylint: disable=empty-docstring
        """ """
        if self.viewing:
            setting = context["instance"]
            context["setting"] = setting

            context["has_html_template"] = self.email_handler.get_auto_body_template(
                setting["key"], "html", fallback_key=setting["fallback_key"]
            )
            context["has_txt_template"] = self.email_handler.get_auto_body_template(
                setting["key"], "txt", fallback_key=setting["fallback_key"]
            )

        return super().render_to_response(template, context)

    def preview(self):
        """
        View for showing a rendered preview of a given email template.

        This will render the email template according to the "mode"
        requested - i.e. HTML or TXT.
        """
        key = self.request.matchdict["key"]
        setting = self.email_handler.get_email_setting(key)
        context = setting.sample_data()
        mode = self.request.params.get("mode", "html")

        if mode == "txt":
            body = self.email_handler.get_auto_txt_body(
                key, context, fallback_key=setting.fallback_key
            )
            self.request.response.content_type = "text/plain"

        else:  # html
            body = self.email_handler.get_auto_html_body(
                key, context, fallback_key=setting.fallback_key
            )

        self.request.response.text = body
        return self.request.response

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """
        cls._email_defaults(config)
        cls._defaults(config)

    @classmethod
    def _email_defaults(cls, config):
        """ """
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        model_title_plural = cls.get_model_title_plural()
        instance_url_prefix = cls.get_instance_url_prefix()

        # fix permission group
        config.add_wutta_permission_group(
            permission_prefix, model_title_plural, overwrite=False
        )

        # preview
        config.add_route(f"{route_prefix}.preview", f"{instance_url_prefix}/preview")
        config.add_view(
            cls,
            attr="preview",
            route_name=f"{route_prefix}.preview",
            permission=f"{permission_prefix}.view",
        )


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    base = globals()

    EmailSettingView = kwargs.get(  # pylint: disable=invalid-name,redefined-outer-name
        "EmailSettingView", base["EmailSettingView"]
    )
    EmailSettingView.defaults(config)


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)



# -*- coding: utf-8; -*-

from unittest.mock import patch

from sqlalchemy import orm

import colander

from wuttaweb.views import roles as mod
from wuttaweb.grids import Grid
from wuttaweb.forms.schema import RoleRef
from wuttaweb.testing import WebTestCase


class TestRoleView(WebTestCase):

    def make_view(self):
        return mod.RoleView(self.request)

    def test_includeme(self):
        self.pyramid_config.include("wuttaweb.views.roles")

    def test_get_query(self):
        view = self.make_view()
        query = view.get_query(session=self.session)
        self.assertIsInstance(query, orm.Query)

    def test_configure_grid(self):
        model = self.app.model
        view = self.make_view()
        grid = view.make_grid(model_class=model.Role)
        self.assertFalse(grid.is_linked("name"))
        view.configure_grid(grid)
        self.assertTrue(grid.is_linked("name"))

    def test_is_editable(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        self.session.commit()
        view = self.make_view()

        admin = auth.get_role_administrator(self.session)
        authed = auth.get_role_authenticated(self.session)
        anon = auth.get_role_anonymous(self.session)

        # editable by default
        self.assertTrue(view.is_editable(blokes))

        # built-in roles not editable by default
        self.assertFalse(view.is_editable(admin))
        self.assertFalse(view.is_editable(authed))
        self.assertFalse(view.is_editable(anon))

        # reset
        del self.request.user_permissions

        barney = model.User(username="barney")
        self.session.add(barney)
        barney.roles.append(blokes)
        auth.grant_permission(blokes, "roles.edit_builtin")
        self.session.commit()

        # user with perms can edit *some* built-in
        self.request.user = barney
        self.assertTrue(view.is_editable(authed))
        self.assertTrue(view.is_editable(anon))
        # nb. not this one yet
        self.assertFalse(view.is_editable(admin))

    def test_is_deletable(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        self.session.commit()
        view = self.make_view()

        # deletable by default
        self.assertTrue(view.is_deletable(blokes))

        # built-in roles not deletable
        self.assertFalse(view.is_deletable(auth.get_role_administrator(self.session)))
        self.assertFalse(view.is_deletable(auth.get_role_authenticated(self.session)))
        self.assertFalse(view.is_deletable(auth.get_role_anonymous(self.session)))

    def test_configure_form(self):
        model = self.app.model
        role = model.Role(name="Foo")
        view = self.make_view()
        form = view.make_form(model_instance=role)
        self.assertNotIn("name", form.validators)
        view.configure_form(form)
        self.assertIsNotNone(form.validators["name"])

    def test_make_users_grid(self):
        self.pyramid_config.add_route("users.view", "/users/{uuid}/view")
        self.pyramid_config.add_route("users.edit", "/users/{uuid}/edit")
        model = self.app.model
        view = self.make_view()
        role = model.Role(name="Manager")
        self.session.add(role)
        user = model.User(username="freddie")
        user.roles.append(role)
        self.session.add(user)
        self.session.commit()

        # basic
        grid = view.make_users_grid(role)
        self.assertIsInstance(grid, Grid)
        self.assertFalse(grid.linked_columns)
        self.assertFalse(grid.actions)

        # view + edit actions
        with patch.object(self.request, "is_root", new=True):
            grid = view.make_users_grid(role)
            self.assertIsInstance(grid, Grid)
            self.assertIn("person", grid.linked_columns)
            self.assertIn("username", grid.linked_columns)
            self.assertEqual(len(grid.actions), 2)
            self.assertEqual(grid.actions[0].key, "view")
            self.assertEqual(grid.actions[1].key, "edit")

            # render grid to ensure coverage for link urls
            grid.render_vue_template()

    def test_unique_name(self):
        model = self.app.model
        view = self.make_view()

        role = model.Role(name="Foo")
        self.session.add(role)
        self.session.commit()

        with patch.object(mod, "Session", return_value=self.session):

            # invalid if same name in data
            node = colander.SchemaNode(colander.String(), name="name")
            self.assertRaises(colander.Invalid, view.unique_name, node, "Foo")

            # but not if name belongs to current role
            view.editing = True
            self.request.matchdict = {"uuid": role.uuid}
            node = colander.SchemaNode(colander.String(), name="name")
            self.assertIsNone(view.unique_name(node, "Foo"))

    def get_permissions(self):
        return {
            "widgets": {
                "label": "Widgets",
                "perms": {
                    "widgets.list": {
                        "label": "List widgets",
                    },
                    "widgets.polish": {
                        "label": "Polish the widgets",
                    },
                    "widgets.view": {
                        "label": "View widget",
                    },
                },
            },
        }

    def test_get_available_permissions(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        blokes = model.Role(name="Blokes")
        auth.grant_permission(blokes, "widgets.list")
        self.session.add(blokes)
        barney = model.User(username="barney")
        barney.roles.append(blokes)
        self.session.add(barney)
        self.session.commit()
        view = self.make_view()
        all_perms = self.get_permissions()
        self.request.registry.settings["wutta_permissions"] = all_perms

        def has_perm(perm):
            if perm == "widgets.list":
                return True
            return False

        with patch.object(self.request, "has_perm", new=has_perm, create=True):

            # sanity check; current request has 1 perm
            self.assertTrue(self.request.has_perm("widgets.list"))
            self.assertFalse(self.request.has_perm("widgets.polish"))
            self.assertFalse(self.request.has_perm("widgets.view"))

            # when editing, user sees only the 1 perm
            with patch.object(view, "editing", new=True):
                perms = view.get_available_permissions()
                self.assertEqual(list(perms), ["widgets"])
                self.assertEqual(list(perms["widgets"]["perms"]), ["widgets.list"])

            # but when viewing, same user sees all perms
            with patch.object(view, "viewing", new=True):
                perms = view.get_available_permissions()
                self.assertEqual(list(perms), ["widgets"])
                self.assertEqual(
                    list(perms["widgets"]["perms"]),
                    ["widgets.list", "widgets.polish", "widgets.view"],
                )

            # also, when admin user is editing, sees all perms
            self.request.is_admin = True
            with patch.object(view, "editing", new=True):
                perms = view.get_available_permissions()
                self.assertEqual(list(perms), ["widgets"])
                self.assertEqual(
                    list(perms["widgets"]["perms"]),
                    ["widgets.list", "widgets.polish", "widgets.view"],
                )

    def test_objectify(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        barney = model.User(username="barney")
        barney.roles.append(blokes)
        self.session.add(barney)
        self.session.commit()
        view = self.make_view()
        permissions = self.get_permissions()

        # sanity check, role has just 1 perm
        auth.grant_permission(blokes, "widgets.list")
        self.session.commit()
        self.assertEqual(blokes.permissions, ["widgets.list"])

        # form can update role perms
        view.editing = True
        self.request.matchdict = {"uuid": blokes.uuid}
        with patch.object(view, "get_available_permissions", return_value=permissions):
            form = view.make_model_form(model_instance=blokes)
            form.validated = {
                "name": "Blokes",
                "permissions": {"widgets.list", "widgets.polish", "widgets.view"},
            }
            role = view.objectify(form)
        self.session.commit()
        self.assertIs(role, blokes)
        self.assertEqual(
            blokes.permissions, ["widgets.list", "widgets.polish", "widgets.view"]
        )

    def test_update_permissions(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        blokes = model.Role(name="Blokes")
        auth.grant_permission(blokes, "widgets.list")
        self.session.add(blokes)
        barney = model.User(username="barney")
        barney.roles.append(blokes)
        self.session.add(barney)
        self.session.commit()
        view = self.make_view()
        permissions = self.get_permissions()

        with patch.object(view, "get_available_permissions", return_value=permissions):

            # no error if data is missing perms
            form = view.make_model_form(model_instance=blokes)
            form.validated = {"name": "BloX"}
            role = view.objectify(form)
            self.session.commit()
            self.assertIs(role, blokes)
            self.assertEqual(blokes.name, "BloX")

            # sanity check, role has just 1 perm
            self.assertEqual(blokes.permissions, ["widgets.list"])

            # role perms are updated
            form = view.make_model_form(model_instance=blokes)
            form.validated = {
                "name": "Blokes",
                "permissions": {"widgets.polish", "widgets.view"},
            }
            role = view.objectify(form)
            self.session.commit()
            self.assertIs(role, blokes)
            self.assertEqual(blokes.permissions, ["widgets.polish", "widgets.view"])


class TestPermissionView(WebTestCase):

    def make_view(self):
        return mod.PermissionView(self.request)

    def test_get_query(self):
        view = self.make_view()
        query = view.get_query(session=self.session)
        self.assertIsInstance(query, orm.Query)

    def test_configure_grid(self):
        model = self.app.model
        view = self.make_view()
        grid = view.make_grid(model_class=model.Permission)
        self.assertFalse(grid.is_linked("role"))
        view.configure_grid(grid)
        self.assertTrue(grid.is_linked("role"))

    def test_configure_form(self):
        model = self.app.model
        role = model.Role(name="Foo")
        perm = model.Permission(role=role, permission="whatever")
        view = self.make_view()
        form = view.make_form(model_instance=perm)
        self.assertIsNone(form.schema)
        view.configure_form(form)
        schema = form.get_schema()
        self.assertIsInstance(schema["role"].typ, RoleRef)



# -*- coding: utf-8; -*-

import datetime
import decimal
from unittest import TestCase
from unittest.mock import patch

import colander
from pyramid import testing

from sqlalchemy import orm

from wuttjamaican.conf import WuttaConfig
from wuttjamaican.util import get_timezone_by_name
from wuttjamaican.testing import DataTestCase
from wuttaweb.forms import schema as mod
from wuttaweb.forms import widgets
from wuttaweb.testing import WebTestCase


class TestWuttaDateTime(WebTestCase):

    def test_serialize(self):
        tzlocal = get_timezone_by_name("America/Los_Angeles")
        with patch.object(self.app, "get_timezone", return_value=tzlocal):
            typ = mod.WuttaDateTime()
            node = colander.SchemaNode(
                typ, widget=widgets.WuttaDateTimeWidget(self.request)
            )

            # null
            self.assertIs(typ.serialize(node, colander.null), colander.null)
            self.assertIs(typ.serialize(node, None), colander.null)
            self.assertIs(typ.serialize(node, ""), colander.null)

            # naive, UTC
            result = typ.serialize(node, datetime.datetime(2024, 12, 11, 22, 33))
            self.assertEqual(result, "2024-12-11T14:33:00-08:00")

            # aware, UTC
            result = typ.serialize(
                node,
                datetime.datetime(2024, 12, 11, 22, 33, tzinfo=datetime.timezone.utc),
            )
            self.assertEqual(result, "2024-12-11T14:33:00-08:00")

            # aware, local
            result = typ.serialize(
                node,
                datetime.datetime(2024, 12, 11, 14, 33, tzinfo=tzlocal),
            )
            self.assertEqual(result, "2024-12-11T14:33:00-08:00")

            # custom format
            typ = mod.WuttaDateTime(format="%Y-%m-%d %I:%M %p")
            node = colander.SchemaNode(
                typ, widget=widgets.WuttaDateTimeWidget(self.request)
            )
            result = typ.serialize(
                node,
                datetime.datetime(2024, 12, 11, 22, 33, tzinfo=datetime.timezone.utc),
            )
            self.assertEqual(result, "2024-12-11 02:33 PM")

            # missing widget/request/config
            typ = mod.WuttaDateTime()
            node = colander.SchemaNode(typ)
            result = typ.serialize(node, datetime.datetime(2024, 12, 11, 22, 33))
            # nb. not possible to know which timezone is system-local
            self.assertTrue(result.startswith("2024-12-"))

    def test_deserialize(self):
        tzlocal = get_timezone_by_name("America/Los_Angeles")
        with patch.object(self.app, "get_timezone", return_value=tzlocal):
            typ = mod.WuttaDateTime()
            node = colander.SchemaNode(
                typ, widget=widgets.WuttaDateTimeWidget(self.request)
            )

            # null
            self.assertIs(typ.deserialize(node, colander.null), colander.null)
            self.assertIs(typ.deserialize(node, None), colander.null)
            self.assertIs(typ.deserialize(node, ""), colander.null)

            # format #1
            result = typ.deserialize(node, "2024-12-11T22:33:00")
            self.assertIsInstance(result, datetime.datetime)
            self.assertEqual(
                result, datetime.datetime(2024, 12, 12, 6, 33, tzinfo=None)
            )

            # format #2
            result = typ.deserialize(node, "2024-12-11T10:33 PM")
            self.assertIsInstance(result, datetime.datetime)
            self.assertEqual(
                result, datetime.datetime(2024, 12, 12, 6, 33, tzinfo=None)
            )

            # invalid
            self.assertRaises(colander.Invalid, typ.deserialize, node, "bogus")


class TestObjectNode(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_dictify(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")

        # unsupported type is converted to string
        node = mod.ObjectNode(colander.String())
        value = node.dictify(person)
        self.assertEqual(value, "Betty Boop")

        # but supported type can dictify
        node = mod.ObjectNode(mod.PersonRef(self.request))
        value = node.dictify(person)
        self.assertIs(value, person)

    def test_objectify(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")

        # unsupported type raises error
        node = mod.ObjectNode(colander.String())
        self.assertRaises(NotImplementedError, node.objectify, person)

        # but supported type can objectify
        node = mod.ObjectNode(mod.PersonRef(self.request))
        value = node.objectify(person)
        self.assertIs(value, person)


class TestWuttaEnum(WebTestCase):

    def test_widget_maker(self):
        enum = self.app.enum
        typ = mod.WuttaEnum(self.request, enum.UpgradeStatus)
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.SelectWidget)


MOCK_STATUS_ONE = 1
MOCK_STATUS_TWO = 2
MOCK_STATUS = {
    MOCK_STATUS_ONE: "one",
    MOCK_STATUS_TWO: "two",
}


class TestWuttaDictEnum(WebTestCase):

    def test_serialize(self):

        # null_value is empty string by default
        typ = mod.WuttaDictEnum(self.request, MOCK_STATUS)
        node = colander.SchemaNode(typ)
        self.assertEqual(typ.serialize(node, 1), "1")
        self.assertEqual(typ.serialize(node, colander.null), "")

        # but can override if needed (?)
        typ = mod.WuttaDictEnum(self.request, MOCK_STATUS, null_value="nope")
        node = colander.SchemaNode(typ)
        self.assertEqual(typ.serialize(node, 1), "1")
        self.assertEqual(typ.serialize(node, colander.null), "nope")

    def test_widget_maker(self):
        typ = mod.WuttaDictEnum(self.request, MOCK_STATUS)
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.SelectWidget)
        self.assertEqual(
            widget.values,
            [
                (1, "one"),
                (2, "two"),
            ],
        )


class TestWuttaMoney(WebTestCase):

    def test_widget_maker(self):
        enum = self.app.enum

        # default scale
        typ = mod.WuttaMoney(self.request)
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.WuttaMoneyInputWidget)
        self.assertEqual(widget.scale, 2)

        # custom scale
        typ = mod.WuttaMoney(self.request, scale=4)
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.WuttaMoneyInputWidget)
        self.assertEqual(widget.scale, 4)


class TestWuttaList(WebTestCase):

    def test_constructor(self):
        node = colander.SchemaNode(mod.WuttaList(self.request))
        typ = node.typ
        self.assertIs(typ.request, self.request)
        self.assertIs(typ.config, self.config)
        self.assertIs(typ.app, self.app)


class TestWuttaSet(WebTestCase):

    def test_constructor(self):
        node = colander.SchemaNode(mod.WuttaSet(self.request))
        typ = node.typ
        self.assertIs(typ.request, self.request)
        self.assertIs(typ.config, self.config)
        self.assertIs(typ.app, self.app)


class TestWuttaQuantity(WebTestCase):

    def test_serialize(self):
        node = colander.SchemaNode(mod.WuttaQuantity(self.request))
        typ = node.typ

        # null
        result = typ.serialize(node, colander.null)
        self.assertIs(result, colander.null)
        result = typ.serialize(node, None)
        self.assertIs(result, colander.null)

        # quantity
        result = typ.serialize(node, 42)
        self.assertEqual(result, "42")
        result = typ.serialize(node, 42.00)
        self.assertEqual(result, "42")
        result = typ.serialize(node, decimal.Decimal("42.00"))
        self.assertEqual(result, "42")
        result = typ.serialize(node, 42.13)
        self.assertEqual(result, "42.13")


class TestObjectRef(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_empty_option(self):

        # null by default
        typ = mod.ObjectRef(self.request)
        self.assertIsNone(typ.empty_option)

        # passing true yields default empty option
        typ = mod.ObjectRef(self.request, empty_option=True)
        self.assertEqual(typ.empty_option, ("", "(none)"))

        # can set explicitly
        typ = mod.ObjectRef(self.request, empty_option=("foo", "bar"))
        self.assertEqual(typ.empty_option, ("foo", "bar"))

        # can set just a label
        typ = mod.ObjectRef(self.request, empty_option="(empty)")
        self.assertEqual(typ.empty_option, ("", "(empty)"))

    def test_model_class(self):
        typ = mod.ObjectRef(self.request)
        self.assertRaises(NotImplementedError, getattr, typ, "model_class")

    def test_serialize(self):
        model = self.app.model
        node = colander.SchemaNode(colander.String())

        # null
        typ = mod.ObjectRef(self.request)
        value = typ.serialize(node, colander.null)
        self.assertIs(value, colander.null)
        value = typ.serialize(node, None)
        self.assertIs(value, colander.null)

        # model instance
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()
        self.assertIsNotNone(person.uuid)
        typ = mod.ObjectRef(self.request)
        value = typ.serialize(node, person)
        self.assertEqual(value, person.uuid.hex)

        # null w/ empty option
        typ = mod.ObjectRef(self.request, empty_option=("bad", "BAD"))
        value = typ.serialize(node, colander.null)
        self.assertEqual(value, "bad")

    def test_deserialize(self):
        model = self.app.model
        node = colander.SchemaNode(colander.String())

        # null
        typ = mod.ObjectRef(self.request)
        value = typ.deserialize(node, colander.null)
        self.assertIs(value, colander.null)

        # model instance
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()
        self.assertIsNotNone(person.uuid)
        with patch.object(mod.ObjectRef, "model_class", new=model.Person):
            with patch.object(mod, "Session", return_value=self.session):
                typ = mod.ObjectRef(self.request)
                value = typ.deserialize(node, person.uuid)
                self.assertIs(value, person)

    def test_dictify(self):
        model = self.app.model
        node = colander.SchemaNode(colander.String())

        # model instance
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()
        self.assertIsNotNone(person.uuid)
        typ = mod.ObjectRef(self.request)
        value = typ.dictify(person)
        self.assertIs(value, person)

    def test_objectify(self):
        model = self.app.model
        node = colander.SchemaNode(colander.String())

        # null
        typ = mod.ObjectRef(self.request)
        value = typ.objectify(None)
        self.assertIsNone(value)

        with patch.object(mod, "Session", return_value=self.session):

            # model instance
            person = model.Person(full_name="Betty Boop")
            self.session.add(person)
            self.session.commit()
            self.assertIsNotNone(person.uuid)
            with patch.object(mod.ObjectRef, "model_class", new=model.Person):

                # can specify as uuid
                typ = mod.ObjectRef(self.request)
                value = typ.objectify(person.uuid)
                self.assertIs(value, person)

                # or can specify object proper
                typ = mod.ObjectRef(self.request)
                value = typ.objectify(person)
                self.assertIs(value, person)

            # error if not found
            with patch.object(mod.ObjectRef, "model_class", new=model.Person):
                typ = mod.ObjectRef(self.request)
                self.assertRaises(ValueError, typ.objectify, "WRONG-UUID")

    def test_get_query(self):
        model = self.app.model
        with patch.object(mod.ObjectRef, "model_class", new=model.Person):
            with patch.object(mod, "Session", return_value=self.session):
                typ = mod.ObjectRef(self.request)
                query = typ.get_query()
                self.assertIsInstance(query, orm.Query)

    def test_sort_query(self):
        model = self.app.model
        with patch.object(mod.ObjectRef, "model_class", new=model.Person):
            with patch.object(mod, "Session", return_value=self.session):
                typ = mod.ObjectRef(self.request)
                query = typ.get_query()
                sorted_query = typ.sort_query(query)
                self.assertIs(sorted_query, query)

    def test_widget_maker(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()

        # basic
        with patch.object(mod.ObjectRef, "model_class", new=model.Person):
            with patch.object(mod, "Session", return_value=self.session):
                typ = mod.ObjectRef(self.request)
                widget = typ.widget_maker()
                self.assertEqual(len(widget.values), 1)
                self.assertEqual(widget.values[0][1], "Betty Boop")

        # empty option
        with patch.object(mod.ObjectRef, "model_class", new=model.Person):
            with patch.object(mod, "Session", return_value=self.session):
                typ = mod.ObjectRef(self.request, empty_option=True)
                widget = typ.widget_maker()
                self.assertEqual(len(widget.values), 2)
                self.assertEqual(widget.values[0][1], "(none)")
                self.assertEqual(widget.values[1][1], "Betty Boop")


class TestPersonRef(WebTestCase):

    def test_sort_query(self):
        with patch.object(mod, "Session", return_value=self.session):
            typ = mod.PersonRef(self.request)
            query = typ.get_query()
            self.assertIsInstance(query, orm.Query)
            sorted_query = typ.sort_query(query)
            self.assertIsInstance(sorted_query, orm.Query)
            self.assertIsNot(sorted_query, query)

    def test_get_object_url(self):
        self.pyramid_config.add_route("people.view", "/people/{uuid}")
        model = self.app.model
        with patch.object(mod, "Session", return_value=self.session):
            typ = mod.PersonRef(self.request)

            person = model.Person(full_name="Barney Rubble")
            self.session.add(person)
            self.session.commit()

            url = typ.get_object_url(person)
            self.assertIsNotNone(url)
            self.assertIn(f"/people/{person.uuid}", url)


class TestRoleRef(WebTestCase):

    def test_sort_query(self):
        with patch.object(mod, "Session", return_value=self.session):
            typ = mod.RoleRef(self.request)
            query = typ.get_query()
            self.assertIsInstance(query, orm.Query)
            sorted_query = typ.sort_query(query)
            self.assertIsInstance(sorted_query, orm.Query)
            self.assertIsNot(sorted_query, query)

    def test_get_object_url(self):
        self.pyramid_config.add_route("roles.view", "/roles/{uuid}")
        model = self.app.model
        with patch.object(mod, "Session", return_value=self.session):
            typ = mod.RoleRef(self.request)

            role = model.Role(name="Manager")
            self.session.add(role)
            self.session.commit()

            url = typ.get_object_url(role)
            self.assertIsNotNone(url)
            self.assertIn(f"/roles/{role.uuid}", url)


class TestUserRef(WebTestCase):

    def test_sort_query(self):
        with patch.object(mod, "Session", return_value=self.session):
            typ = mod.UserRef(self.request)
            query = typ.get_query()
            self.assertIsInstance(query, orm.Query)
            sorted_query = typ.sort_query(query)
            self.assertIsInstance(sorted_query, orm.Query)
            self.assertIsNot(sorted_query, query)

    def test_get_object_url(self):
        self.pyramid_config.add_route("users.view", "/users/{uuid}")
        model = self.app.model
        with patch.object(mod, "Session", return_value=self.session):
            typ = mod.UserRef(self.request)

            user = model.User(username="barney")
            self.session.add(user)
            self.session.commit()

            url = typ.get_object_url(user)
            self.assertIsNotNone(url)
            self.assertIn(f"/users/{user.uuid}", url)


class TestRoleRefs(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_widget_maker(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(self.session)
        authed = auth.get_role_authenticated(self.session)
        anon = auth.get_role_anonymous(self.session)
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        self.session.commit()

        with patch.object(mod, "Session", return_value=self.session):

            # with root access, default values include: admin, blokes
            self.request.is_root = True
            typ = mod.RoleRefs(self.request)
            widget = typ.widget_maker()
            self.assertEqual(len(widget.values), 2)
            self.assertEqual(widget.values[0][1], "Administrator")
            self.assertEqual(widget.values[1][1], "Blokes")

            # without root, default values include: blokes
            self.request.is_root = False
            typ = mod.RoleRefs(self.request)
            widget = typ.widget_maker()
            self.assertEqual(len(widget.values), 1)
            self.assertEqual(widget.values[0][1], "Blokes")


class TestPermissions(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_widget_maker(self):

        # no supported permissions
        permissions = {}
        typ = mod.Permissions(self.request, permissions)
        widget = typ.widget_maker()
        self.assertEqual(len(widget.values), 0)

        # supported permissions are morphed to values
        permissions = {
            "widgets": {
                "label": "Widgets",
                "perms": {
                    "widgets.polish": {
                        "label": "Polish the widgets",
                    },
                },
            },
        }
        typ = mod.Permissions(self.request, permissions)
        widget = typ.widget_maker()
        self.assertEqual(len(widget.values), 1)
        self.assertEqual(widget.values[0], ("widgets.polish", "Polish the widgets"))


class TestFileDownload(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_widget_maker(self):

        # sanity / coverage check
        typ = mod.FileDownload(self.request, url="/foo")
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.FileDownloadWidget)
        self.assertEqual(widget.url, "/foo")


class TestEmailRecipients(TestCase):

    def test_serialize(self):
        typ = mod.EmailRecipients()
        node = colander.SchemaNode(typ)

        recips = [
            "alice@example.com",
            "bob@example.com",
        ]
        recips_str = ", ".join(recips)

        # values
        result = typ.serialize(node, recips_str)
        self.assertEqual(result, "\n".join(recips))

        # null
        result = typ.serialize(node, colander.null)
        self.assertIs(result, colander.null)

    def test_deserialize(self):
        typ = mod.EmailRecipients()
        node = colander.SchemaNode(typ)

        recips = [
            "alice@example.com",
            "bob@example.com",
        ]
        recips_str = ", ".join(recips)

        # values
        result = typ.deserialize(node, recips_str)
        self.assertEqual(result, recips_str)

        # null
        result = typ.deserialize(node, colander.null)
        self.assertIs(result, colander.null)

    def test_widget_maker(self):
        typ = mod.EmailRecipients()
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.EmailRecipientsWidget)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2026 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Common Views
"""

import logging

import colander

from wuttaweb.views import View
from wuttaweb.forms import widgets
from wuttaweb.db import Session
from wuttaweb.util import set_app_theme


log = logging.getLogger(__name__)


class CommonView(View):
    """
    Common views shared by all apps.
    """

    def home(self, session=None):
        """
        Home page view.

        Template: ``/home.mako``

        This is normally the view shown when a user navigates to the
        root URL for the web app.
        """
        # pylint: disable=duplicate-code
        model = self.app.model
        session = session or Session()

        # nb. redirect to /setup if no users exist
        user = session.query(model.User).first()
        if not user:
            return self.redirect(self.request.route_url("setup"))
        # pylint: enable=duplicate-code

        # maybe auto-redirect anons to login
        if not self.request.user:
            if self.config.get_bool("wuttaweb.home_redirect_to_login"):
                return self.redirect(self.request.route_url("login"))

        return {
            "index_title": self.app.get_title(),
        }

    def forbidden_view(self):
        """
        This view is shown when a request triggers a 403 Forbidden error.

        Template: ``/forbidden.mako``
        """
        return {"index_title": self.app.get_title()}

    def notfound_view(self):
        """
        This view is shown when a request triggers a 404 Not Found error.

        Template: ``/notfound.mako``
        """
        return {"index_title": self.app.get_title()}

    def feedback(self):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        session = Session()

        # validate form
        schema = self.feedback_make_schema()
        form = self.make_form(schema=schema)
        if not form.validate():
            # TODO: native Form class should better expose error(s)
            dform = form.get_deform()
            return {"error": str(dform.error)}

        # build email template context
        context = dict(form.validated)
        if context["user_uuid"]:
            context["user"] = session.get(model.User, context["user_uuid"])
            context["user_url"] = self.request.route_url(
                "users.view", uuid=context["user_uuid"]
            )
        context["client_ip"] = self.request.client_addr

        # send email
        try:
            self.feedback_send(context)
        except Exception as error:  # pylint: disable=broad-exception-caught
            log.warning("failed to send feedback email", exc_info=True)
            return {"error": str(error) or error.__class__.__name__}

        return {"ok": True}

    def feedback_make_schema(self):  # pylint: disable=empty-docstring
        """ """
        schema = colander.Schema()

        schema.add(colander.SchemaNode(colander.String(), name="referrer"))

        schema.add(
            colander.SchemaNode(colander.String(), name="user_uuid", missing=None)
        )

        schema.add(colander.SchemaNode(colander.String(), name="user_name"))

        schema.add(colander.SchemaNode(colander.String(), name="message"))

        return schema

    def feedback_send(self, context):  # pylint: disable=empty-docstring
        """ """
        self.app.send_email("feedback", context)

    def setup(self, session=None):  # pylint: disable=too-many-locals
        """
        View for first-time app setup, to create admin user.

        Template: ``/setup.mako``

        This page is only meant for one-time use.  As such, if the app
        DB contains any users, this page will always redirect to the
        home page.

        However if no users exist yet, this will show a form which may
        be used to create the first admin user.  When finished, user
        will be redirected to the login page.

        .. note::

           As long as there are no users in the DB, both the home and
           login pages will automatically redirect to this one.
        """
        model = self.app.model
        session = session or Session()

        # nb. this view only available until first user is created
        user = session.query(model.User).first()
        if user:
            return self.redirect(self.request.route_url("home"))

        form = self.make_form(
            fields=["username", "password", "first_name", "last_name"],
            show_button_cancel=False,
            show_button_reset=True,
        )
        form.set_widget("password", widgets.CheckedPasswordWidget())
        form.set_required("first_name", False)
        form.set_required("last_name", False)

        if form.validate():
            auth = self.app.get_auth_handler()
            data = form.validated

            # make user
            user = auth.make_user(session=session, username=data["username"])
            auth.set_user_password(user, data["password"])

            # assign admin role
            admin = auth.get_role_administrator(session)
            user.roles.append(admin)
            admin.notes = (
                'users in this role may "become root".\n\n'
                "it's recommended not to grant other perms to this role."
            )

            # initialize built-in roles
            authed = auth.get_role_authenticated(session)
            authed.notes = (
                "this role represents any user who *is* logged in.\n\n"
                "you may grant any perms you like to it."
            )
            anon = auth.get_role_anonymous(session)
            anon.notes = (
                "this role represents any user who is *not* logged in.\n\n"
                "you may grant any perms you like to it."
            )

            # also make "Site Admin" role
            site_admin_perms = [
                "alembic.migrations.list",
                "alembic.migrations.create",
                "alembic.migrations.view",
                "alembic.migrations.delete",
                "alembic.migrations.configure",
                "alembic.dashboard",
                "alembic.migrate",
                "app_tables.list",
                "app_tables.create",
                "app_tables.view",
                "appinfo.list",
                "appinfo.configure",
                "email_settings.list",
                "email_settings.edit",
                "email_settings.view",
                "master_views.list",
                "master_views.create",
                "master_views.configure",
                "app_tables.view",
                "people.list",
                "people.create",
                "people.view",
                "people.edit",
                "people.delete",
                "people.versions",
                "roles.list",
                "roles.create",
                "roles.view",
                "roles.edit",
                "roles.edit_builtin",
                "roles.delete",
                "roles.versions",
                "settings.list",
                "settings.create",
                "settings.view",
                "settings.edit",
                "settings.delete",
                "settings.delete_bulk",
                "upgrades.list",
                "upgrades.create",
                "upgrades.view",
                "upgrades.edit",
                "upgrades.delete",
                "upgrades.execute",
                "upgrades.download",
                "upgrades.configure",
                "users.list",
                "users.create",
                "users.view",
                "users.edit",
                "users.delete",
                "users.versions",
            ]
            admin2 = model.Role(name="Site Admin")
            admin2.notes = (
                'this is the "daily driver" admin role.\n\n'
                "you may grant any perms you like to it."
            )
            session.add(admin2)
            user.roles.append(admin2)
            for perm in site_admin_perms:
                auth.grant_permission(admin2, perm)

            # maybe make person
            if data["first_name"] or data["last_name"]:
                first = data["first_name"]
                last = data["last_name"]
                person = model.Person(
                    first_name=first,
                    last_name=last,
                    full_name=(f"{first} {last}").strip(),
                )
                session.add(person)
                user.person = person

            self.setup_enhance_admin_user(user)

            # send user to /login
            self.request.session.flash("Account created! Please login below.")
            return self.redirect(self.request.route_url("login"))

        return {
            "index_title": self.app.get_title(),
            "form": form,
        }

    def setup_enhance_admin_user(self, user):
        """
        Further "enhance" the initial admin user when it is first created.

        This does nothing by default; subclass can override if needed.

        :param user: New admin
           :class:`~wuttjamaican:wuttjamaican.db.model.auth.User`
           which was just created as part of initial setup.
        """

    def change_theme(self):
        """
        This view will set the global app theme, then redirect back to
        the referring page.
        """
        theme = self.request.params.get("theme")
        if theme:
            try:
                set_app_theme(self.request, theme, session=Session())
            except Exception as error:  # pylint: disable=broad-exception-caught
                error = self.app.render_error(error)
                self.request.session.flash(f"Failed to set theme: {error}", "error")
        referrer = self.request.params.get("referrer") or self.request.get_referrer()
        return self.redirect(referrer)

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """
        cls._defaults(config)

    @classmethod
    def _defaults(cls, config):

        config.add_wutta_permission_group("common", "(General)", overwrite=False)

        # home page
        config.add_route("home", "/")
        config.add_view(cls, attr="home", route_name="home", renderer="/home.mako")

        # forbidden
        config.add_forbidden_view(
            cls, attr="forbidden_view", renderer="/forbidden.mako"
        )

        # notfound
        # nb. also, auto-correct URLs which require trailing slash
        config.add_notfound_view(
            cls, attr="notfound_view", append_slash=True, renderer="/notfound.mako"
        )

        # feedback
        config.add_route("feedback", "/feedback", request_method="POST")
        config.add_view(
            cls,
            attr="feedback",
            route_name="feedback",
            permission="common.feedback",
            renderer="json",
        )
        config.add_wutta_permission(
            "common", "common.feedback", "Send a feedback message"
        )

        # setup
        config.add_route("setup", "/setup")
        config.add_view(cls, attr="setup", route_name="setup", renderer="/setup.mako")

        # change theme
        config.add_route("change_theme", "/change-theme", request_method="POST")
        config.add_view(cls, attr="change_theme", route_name="change_theme")
        config.add_wutta_permission(
            "common", "common.change_theme", "Change the global app theme"
        )


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    base = globals()

    CommonView = kwargs.get(  # pylint: disable=invalid-name,redefined-outer-name
        "CommonView", base["CommonView"]
    )
    CommonView.defaults(config)


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)



# -*- coding: utf-8; -*-

from wuttaweb import diffs as mod
from wuttaweb.testing import WebTestCase, VersionWebTestCase


class TestWebDiff(WebTestCase):

    def make_diff(self, *args, **kwargs):
        return mod.WebDiff(self.config, *args, **kwargs)

    def test_render_html(self):
        old_data = {"foo": "bar"}
        new_data = {"foo": "baz"}
        diff = self.make_diff(old_data, new_data)
        html = diff.render_html()
        self.assertIn("<table", html)
        self.assertIn("<tr>", html)
        self.assertIn("&#39;bar&#39;", html)
        self.assertIn(f'style="background-color: {diff.old_color}"', html)
        self.assertIn("&#39;baz&#39;", html)
        self.assertIn(f'style="background-color: {diff.new_color}"', html)
        self.assertIn("</tr>", html)
        self.assertIn("</table>", html)


class TestVersionDiff(VersionWebTestCase):

    def make_diff(self, *args, **kwargs):
        return mod.VersionDiff(self.config, *args, **kwargs)

    def test_constructor(self):
        import sqlalchemy_continuum as continuum

        model = self.app.model
        user = model.User(username="fred")
        self.session.add(user)
        self.session.commit()
        user.username = "freddie"
        self.session.commit()
        self.session.delete(user)
        self.session.commit()

        txncls = continuum.transaction_class(model.User)
        vercls = continuum.version_class(model.User)
        versions = self.session.query(vercls).order_by(vercls.transaction_id).all()
        self.assertEqual(len(versions), 3)

        version = versions[0]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "create")
        self.assertEqual(
            diff.fields,
            ["active", "person_uuid", "prevent_edit", "username", "uuid"],
        )

        version = versions[1]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "update")
        self.assertEqual(
            diff.fields,
            ["active", "person_uuid", "prevent_edit", "username", "uuid"],
        )

        version = versions[2]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "delete")
        self.assertEqual(
            diff.fields,
            ["active", "person_uuid", "prevent_edit", "username", "uuid"],
        )

    def test_render_version_value(self):
        import sqlalchemy_continuum as continuum

        model = self.app.model
        person = model.Person(full_name="Fred Flintstone")
        self.session.add(person)

        # create, update, delete user
        user = model.User(username="fred", person=person)
        self.session.add(user)
        self.session.commit()
        user.username = "freddie"
        self.session.commit()
        self.session.delete(user)
        self.session.commit()

        txncls = continuum.transaction_class(model.User)
        vercls = continuum.version_class(model.User)
        versions = self.session.query(vercls).order_by(vercls.transaction_id).all()
        self.assertEqual(len(versions), 3)

        # create (1st version)
        version = versions[0]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "create")
        self.assertEqual(diff.render_old_value("username"), "")
        self.assertIn("fred", diff.render_new_value("username"))
        self.assertNotIn("freddie", diff.render_new_value("username"))
        self.assertEqual(diff.render_old_value("person_uuid"), "")
        # rendered person_uuid includes display name
        html = diff.render_new_value("person_uuid")
        self.assertIn(str(person.uuid), html)
        self.assertIn("Fred Flintstone", html)

        # update (2nd version)
        version = versions[1]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "update")
        self.assertIn("fred", diff.render_old_value("username"))
        self.assertNotIn("freddie", diff.render_old_value("username"))
        self.assertIn("freddie", diff.render_new_value("username"))
        # rendered person_uuid includes display name
        html = diff.render_old_value("person_uuid")
        self.assertIn(str(person.uuid), html)
        self.assertIn("Fred Flintstone", html)
        html = diff.render_new_value("person_uuid")
        self.assertIn(str(person.uuid), html)
        self.assertIn("Fred Flintstone", html)

        # delete (3rd version)
        version = versions[2]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "delete")
        self.assertIn("freddie", diff.render_old_value("username"))
        self.assertEqual(diff.render_new_value("username"), "")
        # rendered person_uuid includes display name
        html = diff.render_old_value("person_uuid")
        self.assertIn(str(person.uuid), html)
        self.assertIn("Fred Flintstone", html)
        self.assertEqual(diff.render_new_value("person_uuid"), "")



# -*- coding: utf-8; -*-

import os
import sys
from unittest.mock import patch

from wuttaweb.testing import WebTestCase
from wuttaweb.views import views as mod
from wuttaweb.views.users import UserView


class TestMasterViewView(WebTestCase):

    def make_view(self):
        return mod.MasterViewView(self.request)

    def test_includeme(self):
        self.pyramid_config.include("wuttaweb.views.views")

    def test_get_grid_data(self):
        view = self.make_view()

        # empty by default, since nothing registered in test setup
        data = view.get_grid_data()
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 0)

        # so let's register one and try again
        self.pyramid_config.add_wutta_master_view(UserView)
        data = view.get_grid_data()
        self.assertGreater(len(data), 0)
        master = data[0]
        self.assertIsInstance(master, dict)
        self.assertEqual(master["model_title_plural"], "Users")
        self.assertEqual(master["model_name"], "User")
        self.assertEqual(master["url_prefix"], "/users")

    def test_configure_grid(self):
        self.pyramid_config.add_route("users", "/users/")
        self.pyramid_config.add_wutta_master_view(UserView)
        view = self.make_view()

        # sanity / coverage check
        grid = view.make_grid(
            columns=["model_title_plural", "url_prefix"], data=view.get_grid_data()
        )
        view.configure_grid(grid)

        # nb. must invoke this to exercise the url logic
        grid.get_vue_context()

    def test_get_template_context(self):
        view = self.make_view()
        with patch.object(view, "Session", return_value=self.session):

            # normal view gets no extra context
            context = view.get_template_context({})
            self.assertIsInstance(context, dict)
            self.assertNotIn("app_models", context)
            self.assertNotIn("view_module_dirs", context)
            self.assertNotIn("view_module_dir", context)
            self.assertNotIn("menu_path", context)
            self.assertNotIn("roles", context)
            self.assertNotIn("listing_roles", context)
            self.assertNotIn("creating_roles", context)
            self.assertNotIn("viewing_roles", context)
            self.assertNotIn("editing_roles", context)
            self.assertNotIn("deleting_roles", context)

            # but 'create' view gets extra context
            with patch.object(view, "creating", new=True):
                context = view.get_template_context({})
                self.assertIsInstance(context, dict)
                self.assertIn("app_models", context)
                self.assertIn("view_module_dirs", context)
                self.assertIn("view_module_dir", context)
                self.assertIn("menu_path", context)
                self.assertIn("roles", context)
                self.assertIn("listing_roles", context)
                self.assertIn("creating_roles", context)
                self.assertIn("viewing_roles", context)
                self.assertIn("editing_roles", context)
                self.assertIn("deleting_roles", context)

                # try that again but this time make sure there is only
                # one possibility for view module path, which is auto
                # selected by default
                with patch.object(
                    view, "get_view_module_options", return_value=["wuttaweb.views"]
                ):
                    context = view.get_template_context({})
                    self.assertEqual(context["view_module_dir"], "wuttaweb.views")

    def test_get_view_module_options(self):
        view = self.make_view()

        # register one master view, which should be reflected in options
        self.pyramid_config.add_wutta_master_view(UserView)
        options = view.get_view_module_options()
        self.assertEqual(len(options), 1)
        self.assertEqual(options[0], "wuttaweb.views")

    def test_suggest_details(self):
        view = self.make_view()

        # first test uses model_class
        sample = {
            "action": "suggest_details",
            "model_option": "model_class",
            "model_name": "Person",
        }
        with patch.object(self.request, "json_body", new=sample, create=True):
            result = view.wizard_action()
            self.assertEqual(result["class_file_name"], "people.py")
            self.assertEqual(result["class_name"], "PersonView")
            self.assertEqual(result["model_name"], "Person")
            self.assertEqual(result["model_title"], "Person")
            self.assertEqual(result["model_title_plural"], "People")
            self.assertEqual(result["route_prefix"], "people")
            self.assertEqual(result["permission_prefix"], "people")
            self.assertEqual(result["url_prefix"], "/people")
            self.assertEqual(result["template_prefix"], "/people")
            self.assertIn("grid_columns", result)
            self.assertIsInstance(result["grid_columns"], str)
            self.assertIn("form_fields", result)
            self.assertIsInstance(result["form_fields"], str)

        # second test uses model_name
        sample = {
            "action": "suggest_details",
            "model_option": "model_name",
            "model_name": "acme_brick",
        }
        with patch.object(self.request, "json_body", new=sample, create=True):
            result = view.wizard_action()
            self.assertEqual(result["class_file_name"], "acme_bricks.py")
            self.assertEqual(result["class_name"], "AcmeBrickView")
            self.assertEqual(result["model_name"], "acme_brick")
            self.assertEqual(result["model_title"], "Acme Brick")
            self.assertEqual(result["model_title_plural"], "Acme Bricks")
            self.assertEqual(result["route_prefix"], "acme_bricks")
            self.assertEqual(result["permission_prefix"], "acme_bricks")
            self.assertEqual(result["url_prefix"], "/acme-bricks")
            self.assertEqual(result["template_prefix"], "/acme-bricks")
            self.assertEqual(result["grid_columns"], "")
            self.assertEqual(result["form_fields"], "")

    def test_write_view_file(self):
        view = self.make_view()
        view_file_path = self.write_file("silly_things.py", "")
        wutta_file_path = os.path.join(
            os.path.dirname(sys.modules["wuttaweb.views"].__file__),
            "silly_things.py",
        )
        self.assertEqual(os.path.getsize(view_file_path), 0)

        # first test w/ Upgrade model_class and target file path
        sample = {
            "action": "write_view_file",
            "view_location": None,
            "view_file_path": view_file_path,
            "overwrite": False,
            "class_name": "UpgradeView",
            "model_option": "model_class",
            "model_name": "Upgrade",
            "model_title": "Upgrade",
            "model_title_plural": "Upgrades",
            "route_prefix": "upgrades",
            "permission_prefix": "upgrades",
            "url_prefix": "/upgrades",
            "template_prefix": "/upgrades",
            "listable": True,
            "creatable": True,
            "viewable": True,
            "editable": True,
            "deletable": True,
            "grid_columns": ["description", "created_by"],
            "form_fields": ["description", "created_by"],
        }
        with patch.object(self.request, "json_body", new=sample, create=True):

            # does not overwrite by default
            result = view.wizard_action()
            self.assertIn("error", result)
            self.assertEqual(result["error"], "File already exists")
            self.assertEqual(os.path.getsize(view_file_path), 0)

            # but can overwrite if requested
            with patch.dict(sample, {"overwrite": True}):
                result = view.wizard_action()
                self.assertNotIn("error", result)
                self.assertGreater(os.path.getsize(view_file_path), 1000)
                self.assertEqual(result["view_file_path"], view_file_path)
                self.assertEqual(
                    result["view_module_path"], "poser.web.views.silly_things"
                )

        # reset file
        with open(view_file_path, "wb") as f:
            pass
        self.assertEqual(os.path.getsize(view_file_path), 0)

        # second test w/ silly_thing model_name and target module path
        sample = {
            "action": "write_view_file",
            "view_location": "wuttaweb.views",
            "view_file_name": "silly_things.py",
            "overwrite": False,
            "class_name": "SillyThingView",
            "model_option": "model_name",
            "model_name": "silly_thing",
            "model_title": "Silly Thing",
            "model_title_plural": "Silly Things",
            "route_prefix": "silly_things",
            "permission_prefix": "silly_things",
            "url_prefix": "/silly-things",
            "template_prefix": "/silly-things",
            "listable": True,
            "creatable": True,
            "viewable": True,
            "editable": True,
            "deletable": True,
            "grid_columns": ["id", "name", "description"],
            "form_fields": ["id", "name", "description"],
        }
        with patch.object(self.request, "json_body", new=sample, create=True):

            # file does not yet exist, so will be written
            result = view.wizard_action()
            self.assertNotIn("error", result)
            self.assertEqual(result["view_file_path"], wutta_file_path)
            self.assertGreater(os.path.getsize(wutta_file_path), 1000)
            self.assertEqual(os.path.getsize(view_file_path), 0)
            self.assertEqual(result["view_module_path"], "wuttaweb.views.silly_things")

            # once file exists, will not overwrite by default
            result = view.wizard_action()
            self.assertIn("error", result)
            self.assertEqual(result["error"], "File already exists")
            self.assertEqual(os.path.getsize(view_file_path), 0)

            # reset file
            with open(wutta_file_path, "wb") as f:
                pass
            self.assertEqual(os.path.getsize(wutta_file_path), 0)

            # can still overrwite explicitly
            with patch.dict(sample, {"overwrite": True}):
                result = view.wizard_action()
                self.assertNotIn("error", result)
                self.assertEqual(result["view_file_path"], wutta_file_path)
                self.assertGreater(os.path.getsize(wutta_file_path), 1000)
                self.assertEqual(os.path.getsize(view_file_path), 0)
                self.assertEqual(
                    result["view_module_path"], "wuttaweb.views.silly_things"
                )

            # nb. must be sure to deleta that file!
            os.remove(wutta_file_path)

    def test_check_route(self):
        self.pyramid_config.add_route("people", "/people/")
        view = self.make_view()
        sample = {
            "action": "check_route",
            "route": "people",
        }

        with patch.object(self.request, "json_body", new=sample, create=True):

            # should get url and path
            result = view.wizard_action()
            self.assertEqual(result["url"], "http://example.com/people/")
            self.assertEqual(result["path"], "/people/")
            self.assertNotIn("problem", result)

            # unless we check a bad route
            with patch.dict(sample, {"route": "invalid_nothing_burger"}):
                result = view.wizard_action()
            self.assertIn("problem", result)
            self.assertNotIn("url", result)
            self.assertNotIn("path", result)

    def test_apply_permissions(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(self.session)
        known = auth.get_role_authenticated(self.session)

        manager = model.Role(name="Manager")
        self.session.add(manager)

        worker = model.Role(name="worker")
        self.session.add(worker)

        fred = model.User(username="fred")
        fred.roles.append(manager)
        fred.roles.append(worker)
        self.session.add(fred)

        self.session.commit()

        self.assertFalse(auth.has_permission(self.session, fred, "people.list"))
        self.assertFalse(auth.has_permission(self.session, fred, "people.create"))
        self.assertFalse(auth.has_permission(self.session, fred, "people.view"))
        self.assertFalse(auth.has_permission(self.session, fred, "people.edit"))
        self.assertFalse(auth.has_permission(self.session, fred, "people.delete"))

        view = self.make_view()
        with patch.object(view, "Session", return_value=self.session):

            sample = {
                "action": "apply_permissions",
                "permission_prefix": "people",
                "listing_roles": {known.uuid.hex: True},
                "creating_roles": {worker.uuid.hex: True},
                "viewing_roles": {known.uuid.hex: True},
                "editing_roles": {manager.uuid.hex: True},
                "deleting_roles": {manager.uuid.hex: True},
            }
            with patch.object(self.request, "json_body", new=sample, create=True):

                # nb. empty result is normal
                result = view.wizard_action()
                self.assertEqual(result, {})

                self.assertTrue(auth.has_permission(self.session, fred, "people.list"))
                self.assertTrue(
                    auth.has_permission(self.session, fred, "people.create")
                )
                self.assertTrue(auth.has_permission(self.session, fred, "people.view"))
                self.assertTrue(auth.has_permission(self.session, fred, "people.edit"))
                self.assertTrue(
                    auth.has_permission(self.session, fred, "people.delete")
                )

    def test_wizard_action(self):
        view = self.make_view()

        # missing action
        with patch.object(self.request, "json_body", create=True, new={}):
            result = view.wizard_action()
            self.assertIn("error", result)
            self.assertEqual(result["error"], "Must specify the action to perform.")

        # unknown action
        with patch.object(
            self.request, "json_body", create=True, new={"action": "nothing"}
        ):
            result = view.wizard_action()
            self.assertIn("error", result)
            self.assertEqual(result["error"], "Unknown action requested: nothing")

        # error invoking action
        with patch.object(
            self.request, "json_body", create=True, new={"action": "check_route"}
        ):
            with patch.object(view, "check_route", side_effect=RuntimeError("whoa")):
                result = view.wizard_action()
                self.assertIn("error", result)
                self.assertEqual(result["error"], "Unexpected error occurred: whoa")

    def test_configure(self):
        self.pyramid_config.add_route("home", "/")
        self.pyramid_config.add_route("login", "/auth/login")
        self.pyramid_config.add_route("master_views", "/views/master")
        view = self.make_view()

        # sanity/coverage
        view.configure()



# -*- coding: utf-8; -*-

import os
from unittest.mock import patch

import sqlalchemy as sa
from alembic import command as alembic_command

from wuttjamaican.testing import ConfigTestCase
from wuttjamaican.db.conf import (
    get_alembic_scriptdir,
    check_alembic_current,
    make_alembic_config,
)

import colander
from pyramid.httpexceptions import HTTPNotFound, HTTPFound

from wuttaweb.views import alembic as mod
from wuttaweb.forms import Form
from wuttaweb.testing import WebTestCase
from wuttaweb.forms.widgets import AlembicRevisionWidget


class TestNormalizeRevision(ConfigTestCase):

    def test_basic(self):
        self.config.setdefault("alembic.script_location", "wuttjamaican.db:alembic")
        self.config.setdefault(
            "alembic.version_locations", "wuttjamaican.db:alembic/versions"
        )

        script = get_alembic_scriptdir(self.config)
        head = script.get_heads()[0]
        rev = script.get_revision(head)

        result = mod.normalize_revision(self.config, rev)
        self.assertIsInstance(result, dict)
        self.assertIn("revision", result)
        self.assertEqual(result["revision"], rev.revision)


class TestAlembicMigrationView(WebTestCase):

    def setUp(self):
        super().setUp()
        self.config.setdefault("alembic.script_location", "wuttjamaican.db:alembic")
        self.config.setdefault(
            "alembic.version_locations", "wuttjamaican.db:alembic/versions"
        )

    def make_view(self):
        return mod.AlembicMigrationView(self.request)

    def test_includeme(self):
        self.pyramid_config.include("wuttaweb.views.alembic")

    def test_get_grid_data(self):
        view = self.make_view()
        data = view.get_grid_data()
        self.assertIsInstance(data, list)
        self.assertTrue(data)  # 1+ items
        rev = data[0]
        self.assertIn("revision", rev)
        self.assertIn("down_revision", rev)
        self.assertIn("doc", rev)

    def test_configure_grid(self):
        view = self.make_view()
        grid = view.make_model_grid()
        self.assertIn("revision", grid.searchable_columns)
        self.assertIn("doc", grid.searchable_columns)

    def test_render_is_head(self):
        view = self.make_view()

        # missing field / empty default
        rev = {"revision": "foo"}
        self.assertEqual(view.render_is_head(rev, "is_head", None), "")

        # boolean true
        rev = {"revision": "foo", "is_head": True}
        self.assertEqual(view.render_is_head(rev, "is_head", True), "Yes")

        # boolean false
        rev = {"revision": "foo", "is_head": False}
        self.assertEqual(view.render_is_head(rev, "is_head", False), "")

    def test_get_instance(self):
        view = self.make_view()

        with patch.object(self.request, "matchdict", new={"revision": "fc3a3bcaa069"}):

            rev1 = view.get_instance()
            self.assertIsInstance(rev1, dict)
            self.assertIn("revision", rev1)
            self.assertEqual(rev1["revision"], "fc3a3bcaa069")
            self.assertIn("doc", rev1)

            rev2 = view.get_instance()
            self.assertIs(rev2, rev1)
            self.assertEqual(rev2["revision"], "fc3a3bcaa069")

        view = self.make_view()
        with patch.object(self.request, "matchdict", new={"revision": "invalid"}):
            self.assertRaises(HTTPNotFound, view.get_instance)

    def test_get_instance_title(self):
        view = self.make_view()

        rev = {
            "revision": "fc3a3bcaa069",
            "doc": "init with settings table",
            "branch_labels": "wutta",
        }
        self.assertEqual(
            view.get_instance_title(rev), "(wutta) init with settings table"
        )

        rev = {
            "revision": "fc3a3bcaa069",
            "doc": "init with settings table",
            "branch_labels": "wutta",
            "is_head": True,
        }
        self.assertEqual(
            view.get_instance_title(rev), "(wutta) init with settings table [head]"
        )

    def test_configure_form(self):
        view = self.make_view()

        # sanity / coverage
        with patch.object(self.request, "matchdict", new={"revision": "fc3a3bcaa069"}):
            rev = view.get_instance()
            form = view.make_model_form(rev)
            self.assertIsInstance(form.widgets["down_revision"], AlembicRevisionWidget)

    def test_make_create_form(self):
        self.pyramid_config.add_route("alembic.migrations", "/alembic/migrations/")
        view = self.make_view()

        # sanity / coverage
        form = view.make_create_form()
        self.assertIsInstance(form, Form)
        self.assertIn("branching_option", form)

    def test_validate_revise_branch(self):
        self.pyramid_config.add_route("alembic.migrations", "/alembic/migrations/")
        view = self.make_view()
        form = view.make_create_form()
        schema = form.get_schema()

        # good example
        self.assertIsNone(
            view.validate_revise_branch(
                schema,
                {
                    "branching_option": "revise",
                    "revise_branch": "wutta",
                },
            )
        )

        # branch is required
        self.assertRaises(
            colander.Invalid,
            view.validate_revise_branch,
            schema,
            {
                "branching_option": "revise",
                "revise_branch": None,
            },
        )

    def test_validate_new_branch(self):
        self.pyramid_config.add_route("alembic.migrations", "/alembic/migrations/")
        view = self.make_view()
        form = view.make_create_form()
        schema = form.get_schema()

        # good example
        self.assertIsNone(
            view.validate_revise_branch(
                schema,
                {
                    "branching_option": "new",
                    "new_branch": "poser",
                    "version_location": "wuttjamaican.db:alembic/versions",
                },
            )
        )

        # name is required
        self.assertRaises(
            colander.Invalid,
            view.validate_new_branch,
            schema,
            {
                "branching_option": "new",
                "new_branch": None,
                "version_location": "wuttjamaican.db:alembic/versions",
            },
        )

        # version_location is required
        self.assertRaises(
            colander.Invalid,
            view.validate_new_branch,
            schema,
            {
                "branching_option": "new",
                "new_branch": "poser",
                "version_location": None,
            },
        )

    def test_save_create_form(self):
        self.pyramid_config.add_route("alembic.migrations", "/alembic/migrations/")
        self.pyramid_config.add_route("alembic.dashboard", "/alembic/dashboard")
        view = self.make_view()
        form = view.make_create_form()

        # revise branch
        form.validated = {
            "description": "test revision",
            "autogenerate": False,
            "branching_option": "revise",
            "revise_branch": "wutta",
        }
        revision = view.save_create_form(form)

        # file was saved in wutta dir
        self.assertTrue(
            revision.path.startswith(
                self.app.resource_path("wuttjamaican.db:alembic/versions")
            )
        )

        # get rid of that file!
        os.remove(revision.path)

        # new branch
        form.validated = {
            "description": "test revision",
            "autogenerate": False,
            "branching_option": "new",
            "new_branch": "wuttatest",
            "version_location": "wuttjamaican.db:alembic/versions",
        }
        revision = view.save_create_form(form)

        # file was saved in wutta dir
        self.assertTrue(
            revision.path.startswith(
                self.app.resource_path("wuttjamaican.db:alembic/versions")
            )
        )

        # get rid of that file!
        os.remove(revision.path)

    def test_save_delete_form(self):
        self.pyramid_config.add_route(
            "alembic.migrations.view", "/alembic/migrations/{revision}"
        )
        view = self.make_view()
        alembic = make_alembic_config(self.config)

        # write new empty migration script
        revision = alembic_command.revision(
            alembic,
            head="base",
            branch_label="wuttatest",
            version_path=self.app.resource_path("wuttjamaican.db:alembic/versions"),
            message="test revision",
        )

        # script exists
        self.assertTrue(os.path.exists(revision.path))

        with patch.object(
            self.request, "matchdict", new={"revision": revision.revision}
        ):
            form = view.make_delete_form(revision)
            view.save_delete_form(form)
            # script gone
            self.assertFalse(os.path.exists(revision.path))

    def test_configure(self):
        self.pyramid_config.add_route("home", "/")
        self.pyramid_config.add_route("login", "/auth/login")
        self.pyramid_config.add_route("alembic.migrations", "/alembic/migrations")
        view = self.make_view()

        # sanity/coverage
        view.configure()


class TestAlembicDashboardView(WebTestCase):

    def make_config(self, **kwargs):
        sqlite_path = self.write_file("test.sqlite", "")
        self.sqlite_engine_url = f"sqlite:///{sqlite_path}"

        config_path = self.write_file(
            "test.ini",
            f"""
[wutta.db]
default.url = {self.sqlite_engine_url}

[alembic]
script_location = wuttjamaican.db:alembic
version_locations = wuttjamaican.db:alembic/versions
""",
        )

        return super().make_config([config_path], **kwargs)

    def make_view(self):
        return mod.AlembicDashboardView(self.request)

    def test_dashboard(self):
        self.pyramid_config.add_route(
            "alembic.migrations.view", "/alembic/migrations/{revision}"
        )
        view = self.make_view()
        alembic = make_alembic_config(self.config)

        # tests use MetaData.create_all() instead of migrations for
        # setup, so alembic will assume db is not current
        self.assertFalse(check_alembic_current(self.config, alembic))

        # and to further prove the point, alembic_version table is missing
        self.assertEqual(
            self.session.execute(sa.text("select count(*) from person")).scalar(),
            0,
        )
        self.assertRaises(
            sa.exc.OperationalError,
            self.session.execute,
            sa.text("select count(*) from alembic_version"),
        )

        # therefore dashboard shows db with no heads at first
        context = view.dashboard()
        self.assertIsInstance(context, dict)
        self.assertIn("script_heads", context)
        self.assertEqual(len(context["script_heads"]), 1)
        self.assertIn("db_heads", context)
        self.assertEqual(len(context["db_heads"]), 0)

        # but we can 'stamp' the db as current
        alembic_command.stamp(alembic, "heads")

        # now the alembic_version table exists
        self.assertEqual(
            self.session.execute(
                sa.text("select count(*) from alembic_version")
            ).scalar(),
            1,
        )
        self.assertTrue(check_alembic_current(self.config, alembic))

        # and the dashboard knows about db heads
        context = view.dashboard()
        self.assertEqual(len(context["script_heads"]), 1)
        self.assertEqual(len(context["db_heads"]), 1)

    def test_migrate(self):
        self.pyramid_config.add_route("alembic.dashboard", "/alembic/dashboard")
        view = self.make_view()

        # tell alembic our db is already current
        alembic = make_alembic_config(self.config)
        alembic_command.stamp(alembic, "heads")
        self.assertTrue(check_alembic_current(self.config, alembic))

        # GET request redirects to dashboard w/ no flash
        result = view.migrate()
        self.assertIsInstance(result, HTTPFound)
        self.assertEqual(result.location, "http://example.com/alembic/dashboard")
        self.assertFalse(self.request.session.peek_flash())
        self.assertFalse(self.request.session.peek_flash("error"))

        # POST with no revspec also redirects but w/ flash
        with patch.object(self.request, "method", new="POST"):
            result = view.migrate()
            self.assertIsInstance(result, HTTPFound)
            self.assertEqual(result.location, "http://example.com/alembic/dashboard")
            self.assertFalse(self.request.session.peek_flash())
            self.assertTrue(self.request.session.peek_flash("error"))
            self.assertEqual(
                self.request.session.pop_flash("error"),
                ["You must provide a target revspec."],
            )

        # force downgrade to wutta@-1
        with patch.object(self.request, "method", new="POST"):
            with patch.object(
                self.request,
                "POST",
                new={"direction": "downgrade", "revspec": "wutta@-1"},
            ):
                # nb. this still redirects but w/ different flash
                result = view.migrate()
                self.assertIsInstance(result, HTTPFound)
                self.assertEqual(
                    result.location, "http://example.com/alembic/dashboard"
                )
                self.assertTrue(self.request.session.peek_flash())
                self.assertFalse(self.request.session.peek_flash("error"))
                self.assertEqual(
                    self.request.session.pop_flash(),
                    ["Database has been migrated."],
                )

        # alembic should know our db is no longer current
        self.assertFalse(check_alembic_current(self.config, alembic))

        # force upgrade to heads
        with patch.object(self.request, "method", new="POST"):
            with patch.object(
                self.request,
                "POST",
                new={"revspec": "heads"},
            ):
                # nb. this still redirects but w/ different flash
                result = view.migrate()
                self.assertIsInstance(result, HTTPFound)
                self.assertEqual(
                    result.location, "http://example.com/alembic/dashboard"
                )
                self.assertTrue(self.request.session.peek_flash())
                self.assertFalse(self.request.session.peek_flash("error"))
                self.assertEqual(
                    self.request.session.pop_flash(),
                    ["Database has been migrated."],
                )

        # alembic should know our db is current again
        self.assertTrue(check_alembic_current(self.config, alembic))

        # upgrade to invalid spec (force an error)
        with patch.object(self.request, "method", new="POST"):
            with patch.object(
                self.request,
                "POST",
                new={"revspec": "bad-spec"},
            ):
                # nb. this still redirects but w/ different flash
                result = view.migrate()
                self.assertIsInstance(result, HTTPFound)
                self.assertEqual(
                    result.location, "http://example.com/alembic/dashboard"
                )
                self.assertFalse(self.request.session.peek_flash())
                self.assertTrue(self.request.session.peek_flash("error"))
                [msg] = self.request.session.pop_flash("error")
                self.assertTrue(msg.startswith("Migrate failed: "))



# -*- coding: utf-8; -*-

import os
from unittest.mock import patch

from alembic import command as alembic_command

from wuttjamaican.db.conf import check_alembic_current, make_alembic_config

from wuttaweb.testing import WebTestCase
from wuttaweb.views import tables as mod
from wuttaweb.views.users import UserView


class TestAppTableView(WebTestCase):

    def make_config(self, **kwargs):
        sqlite_path = self.write_file("test.sqlite", "")
        self.sqlite_engine_url = f"sqlite:///{sqlite_path}"

        config_path = self.write_file(
            "test.ini",
            f"""
[wutta.db]
default.url = {self.sqlite_engine_url}

[alembic]
script_location = wuttjamaican.db:alembic
version_locations = wuttjamaican.db:alembic/versions
""",
        )

        return super().make_config([config_path], **kwargs)

    def make_view(self):
        return mod.AppTableView(self.request)

    def test_includeme(self):
        self.pyramid_config.include("wuttaweb.views.tables")

    def test_get_grid_data(self):
        view = self.make_view()
        data = view.get_grid_data()
        self.assertIsInstance(data, list)
        self.assertGreater(len(data), 0)
        table = data[0]
        self.assertIsInstance(table, dict)
        self.assertIn("name", table)
        self.assertIn("schema", table)

    def test_configure_grid(self):
        view = self.make_view()

        # sanity / coverage check
        grid = view.make_grid(columns=["name", "schema"])
        view.configure_grid(grid)

    def test_get_instance(self):
        view = self.make_view()

        with patch.object(self.request, "matchdict", new={"name": "person"}):

            table1 = view.get_instance()
            self.assertIsInstance(table1, dict)
            self.assertIn("name", table1)
            self.assertEqual(table1["name"], "person")
            self.assertIn("schema", table1)

            table2 = view.get_instance()
            self.assertIs(table2, table1)
            self.assertEqual(table2["name"], "person")

    def test_get_instance_title(self):
        view = self.make_view()

        table = {"name": "poser_foo"}
        self.assertEqual(view.get_instance_title(table), "poser_foo")

    def test_configure_form(self):
        view = self.make_view()
        table = {"name": "user", "description": "Represents a user"}

        # no description widget by default
        form = view.make_form(model_instance=table, fields=["name", "description"])
        self.assertNotIn("description", form.widgets)

        # but it gets added when configuring
        view.configure_form(form)
        self.assertIn("description", form.widgets)

    def test_get_xref_buttons(self):
        self.pyramid_config.add_route("users", "/users/")
        self.pyramid_config.add_route("master_views.create", "/views/master/new")
        model = self.app.model
        view = self.make_view()

        # nb. must add this first
        self.pyramid_config.add_wutta_master_view(UserView)

        # should be just one xref button by default
        table = {"name": "user", "model_class": model.User}
        buttons = view.get_xref_buttons(table)
        self.assertEqual(len(buttons), 1)
        button = buttons[0]
        self.assertIn("Users", button)
        self.assertIn("http://example.com/users/", button)

        # unless we have perm to make new master view
        with patch.object(self.request, "is_root", new=True):
            table = {"name": "user", "model_class": model.User}
            buttons = view.get_xref_buttons(table)
            self.assertEqual(len(buttons), 3)
            first, second, third = buttons
            self.assertIn("Users", first)
            self.assertIn("http://example.com/users/", first)
            self.assertEqual(second, "<br />")
            self.assertIn("New Master View", third)
            self.assertIn("http://example.com/views/master/new", third)

    def test_get_row_grid_data(self):
        model = self.app.model
        view = self.make_view()

        table = model.Base.metadata.tables["person"]
        table_dict = {"name": "person", "table": table}

        data = view.get_row_grid_data(table_dict)
        self.assertIsInstance(data, list)
        self.assertGreater(len(data), 4)
        columns = [c["column_name"] for c in data]
        self.assertIn("full_name", columns)
        self.assertIn("first_name", columns)
        self.assertIn("last_name", columns)

    def test_configure_row_grid(self):
        view = self.make_view()

        # sanity / coverage check
        grid = view.make_grid(columns=["column_name", "data_type"])
        view.configure_row_grid(grid)

    def test_render_column_description(self):
        view = self.make_view()

        # nb. first 2 params are igored
        text = view.render_column_description(None, None, "hello world")
        self.assertEqual(text, "hello world")

        text = view.render_column_description(None, None, "")
        self.assertEqual(text, "")
        text = view.render_column_description(None, None, None)
        self.assertEqual(text, "")

        msg = (
            "This is a very long and rambling sentence.  "
            "There is no point to it except that it is long.  "
            "Far too long to be reasonable."
            "I mean I am serious when I say this is simply too long."
        )
        text = view.render_column_description(None, None, msg)
        self.assertNotEqual(text, msg)
        self.assertIn("<span title=", text)

    def test_get_template_context(self):
        view = self.make_view()

        # normal view gets no extra context
        context = view.get_template_context({})
        self.assertIsInstance(context, dict)
        self.assertNotIn("alembic_is_current", context)
        self.assertNotIn("existing_tables", context)
        self.assertNotIn("model_dir", context)
        self.assertNotIn("migration_branch_options", context)
        self.assertNotIn("migration_branch", context)

        # but 'create' view gets extra context
        with patch.object(view, "creating", new=True):
            context = view.get_template_context({})
            self.assertIsInstance(context, dict)
            self.assertIn("alembic_is_current", context)
            self.assertIn("existing_tables", context)
            self.assertIn("model_dir", context)
            self.assertIn("migration_branch_options", context)
            self.assertIn("migration_branch", context)

    def test_write_model_file(self):
        view = self.make_view()
        module_path = self.write_file("widget.py", "")
        self.assertEqual(os.path.getsize(module_path), 0)

        sample = {
            "action": "write_model_file",
            "module_file": module_path,
            "overwrite": False,
            "table_name": "poser_widget",
            "model_name": "PoserWidget",
            "model_title": "Poser Widget",
            "model_title_plural": "Poser Widgets",
            "description": "A widget for Poser",
            "versioned": True,
            "columns": [
                {
                    "name": "uuid",
                    "data_type": {
                        "type": "UUID",
                    },
                    "formatted_data_type": "sa.UUID()",
                    "nullable": False,
                    "description": "primary key",
                    "versioned": "n/a",
                    "relationship": None,
                },
                {
                    "name": "name",
                    "data_type": {
                        "type": "String",
                    },
                    "formatted_data_type": "sa.String(length=100)",
                    "nullable": False,
                    "description": "name of widget",
                    "versioned": True,
                    "relationship": None,
                },
                {
                    "name": "owner_uuid",
                    "data_type": {
                        "type": "_fk_uuid_",
                        "reference": "user",
                    },
                    "formatted_data_type": "sa.UUID()",
                    "nullable": False,
                    "description": "owner of widget",
                    "versioned": True,
                    "relationship": {
                        "name": "owner",
                        "reference_model": "User",
                    },
                },
            ],
        }

        with patch.object(self.request, "json_body", new=sample, create=True):

            # does not overwrite by default
            result = view.wizard_action()
            self.assertIn("error", result)
            self.assertEqual(result["error"], "File already exists")
            self.assertEqual(os.path.getsize(module_path), 0)

            # but it can overwrite if requested
            with patch.dict(sample, {"overwrite": True}):
                result = view.wizard_action()
                self.assertNotIn("error", result)
                self.assertGreater(os.path.getsize(module_path), 500)

    def test_check_model(self):
        view = self.make_view()
        sample = {
            "action": "check_model",
            "model_name": "Person",
        }

        with patch.object(self.request, "json_body", new=sample, create=True):

            # empty result means the model exists
            result = view.wizard_action()
            self.assertEqual(result, {})

            # problem is specified if not
            with patch.dict(sample, {"model_name": "gobbledygook"}):
                result = view.wizard_action()
                self.assertIn("problem", result)
                self.assertEqual(result["problem"], "class not found in app model")

    def test_write_revision_script(self):
        view = self.make_view()
        sample = {
            "action": "write_revision_script",
            "branch": "wutta",
            "message": "just a test",
        }

        # tell alembic our db is already current
        alembic = make_alembic_config(self.config)
        alembic_command.stamp(alembic, "heads")
        self.assertTrue(check_alembic_current(self.config, alembic))

        with patch.object(self.request, "json_body", new=sample, create=True):

            # nb. this writes a real script in the wuttjamaican project
            result = view.wizard_action()
            self.assertIn("script", result)
            outdir = os.path.dirname(result["script"])
            self.assertEqual(
                outdir, self.app.resource_path("wuttjamaican.db:alembic/versions")
            )

            # alembic now thinks we need to upgrade
            self.assertFalse(check_alembic_current(self.config, alembic))

            # must be sure to delete that script
            os.remove(result["script"])

    def test_migrate_db(self):
        view = self.make_view()
        sample = {"action": "migrate_db"}

        # tell alembic our db is already current
        alembic = make_alembic_config(self.config)
        alembic_command.stamp(alembic, "heads")
        self.assertTrue(check_alembic_current(self.config, alembic))

        # force downgrade to wutta@-1
        alembic_command.downgrade(alembic, "wutta@-1")

        # alembic now thinks we need to upgrade
        self.assertFalse(check_alembic_current(self.config, alembic))

        with patch.object(self.request, "json_body", new=sample, create=True):

            # now test our view method; alembic should then know we are current
            view.wizard_action()
            self.assertTrue(check_alembic_current(self.config, alembic))

    def test_check_table(self):
        self.pyramid_config.add_route("app_tables.view", "/tables/app/{name}")
        view = self.make_view()
        sample = {"action": "check_table", "name": "person"}

        with patch.object(view, "Session", return_value=self.session):

            with patch.object(self.request, "json_body", new=sample, create=True):

                # result with URL means the table exists
                result = view.wizard_action()
                self.assertIn("url", result)
                self.assertNotIn("problem", result)

                # problem is specified if not
                with patch.dict(sample, {"name": "gobbledygook"}):
                    result = view.wizard_action()
                self.assertIn("problem", result)
                self.assertEqual(result["problem"], "table does not exist in app model")

    def test_wizard_action(self):
        view = self.make_view()

        # missing action
        with patch.object(self.request, "json_body", create=True, new={}):
            result = view.wizard_action()
            self.assertIn("error", result)
            self.assertEqual(result["error"], "Must specify the action to perform.")

        # unknown action
        with patch.object(
            self.request, "json_body", create=True, new={"action": "nothing"}
        ):
            result = view.wizard_action()
            self.assertIn("error", result)
            self.assertEqual(result["error"], "Unknown action requested: nothing")

        # error invoking action
        with patch.object(
            self.request, "json_body", create=True, new={"action": "check_table"}
        ):
            with patch.object(view, "check_table", side_effect=RuntimeError("whoa")):
                result = view.wizard_action()
                self.assertIn("error", result)
                self.assertEqual(result["error"], "Unexpected error occurred: whoa")
